@@ -4,12 +4,9 @@
  * license that can be found in the LICENSE file.
  */
 
-#include <openssl/ec.h>
-#include <openssl/evp.h>
 #include <openssl/sha.h>
 #include <openssl/x509.h>
 
-#include <string.h>
 #include "fido.h"
 #include "fido/es256.h"
 
@@ -37,6 +34,8 @@ parse_makecred_reply(const cbor_item_t *key, const cbor_item_t *val, void *arg)
 		    &cred->authdata_ext));
 	case 3: /* attestation statement */
 		return (cbor_decode_attstmt(val, &cred->attstmt));
+	case 5: /* large blob key */
+		return (fido_blob_decode(val, &cred->largeblob_key));
 	default: /* ignore */
 		fido_log_debug("%s: cbor type", __func__);
 		return (0);
@@ -48,8 +47,10 @@ fido_dev_make_cred_tx(fido_dev_t *dev, fido_cred_t *cred, const char *pin)
 {
 	fido_blob_t	 f;
 	fido_blob_t	*ecdh = NULL;
+	fido_opt_t	 uv = cred->uv;
 	es256_pk_t	*pk = NULL;
 	cbor_item_t	*argv[9];
+	const uint8_t	 cmd = CTAP_CBOR_MAKECRED;
 	int		 r;
 
 	memset(&f, 0, sizeof(f));
@@ -81,36 +82,38 @@ fido_dev_make_cred_tx(fido_dev_t *dev, fido_cred_t *cred, const char *pin)
 
 	/* extensions */
 	if (cred->ext.mask)
-		if ((argv[5] = cbor_encode_extensions(&cred->ext)) == NULL) {
-			fido_log_debug("%s: cbor_encode_extensions", __func__);
+		if ((argv[5] = cbor_encode_cred_ext(&cred->ext,
+		    &cred->blob)) == NULL) {
+			fido_log_debug("%s: cbor_encode_cred_ext", __func__);
 			r = FIDO_ERR_INTERNAL;
 			goto fail;
 		}
 
-	/* options */
-	if (cred->rk != FIDO_OPT_OMIT || cred->uv != FIDO_OPT_OMIT)
-		if ((argv[6] = cbor_encode_options(cred->rk,
-		    cred->uv)) == NULL) {
-			fido_log_debug("%s: cbor_encode_options", __func__);
-			r = FIDO_ERR_INTERNAL;
-			goto fail;
-		}
-
-	/* pin authentication */
-	if (pin) {
+	/* user verification */
+	if (pin != NULL || (uv == FIDO_OPT_TRUE &&
+	    fido_dev_supports_permissions(dev))) {
 		if ((r = fido_do_ecdh(dev, &pk, &ecdh)) != FIDO_OK) {
 			fido_log_debug("%s: fido_do_ecdh", __func__);
 			goto fail;
 		}
-		if ((r = cbor_add_pin_params(dev, &cred->cdh, pk, ecdh, pin,
-		    &argv[7], &argv[8])) != FIDO_OK) {
-			fido_log_debug("%s: cbor_add_pin_params", __func__);
+		if ((r = cbor_add_uv_params(dev, cmd, &cred->cdh, pk, ecdh,
+		    pin, cred->rp.id, &argv[7], &argv[8])) != FIDO_OK) {
+			fido_log_debug("%s: cbor_add_uv_params", __func__);
 			goto fail;
 		}
+		uv = FIDO_OPT_OMIT;
 	}
 
+	/* options */
+	if (cred->rk != FIDO_OPT_OMIT || uv != FIDO_OPT_OMIT)
+		if ((argv[6] = cbor_encode_cred_opt(cred->rk, uv)) == NULL) {
+			fido_log_debug("%s: cbor_encode_cred_opt", __func__);
+			r = FIDO_ERR_INTERNAL;
+			goto fail;
+		}
+
 	/* framing and transmission */
-	if (cbor_build_frame(CTAP_CBOR_MAKECRED, argv, nitems(argv), &f) < 0 ||
+	if (cbor_build_frame(cmd, argv, nitems(argv), &f) < 0 ||
 	    fido_tx(dev, CTAP_CMD_CBOR, f.ptr, f.len) < 0) {
 		fido_log_debug("%s: fido_tx", __func__);
 		r = FIDO_ERR_TX;
@@ -149,8 +152,7 @@ fido_dev_make_cred_rx(fido_dev_t *dev, fido_cred_t *cred, int ms)
 	}
 
 	if (cred->fmt == NULL || fido_blob_is_empty(&cred->authdata_cbor) ||
-	    fido_blob_is_empty(&cred->attcred.id) ||
-	    fido_blob_is_empty(&cred->attstmt.sig)) {
+	    fido_blob_is_empty(&cred->attcred.id)) {
 		fido_cred_reset_rx(cred);
 		return (FIDO_ERR_INVALID_CBOR);
 	}
@@ -159,7 +161,8 @@ fido_dev_make_cred_rx(fido_dev_t *dev, fido_cred_t *cred, int ms)
 }
 
 static int
-fido_dev_make_cred_wait(fido_dev_t *dev, fido_cred_t *cred, const char *pin, int ms)
+fido_dev_make_cred_wait(fido_dev_t *dev, fido_cred_t *cred, const char *pin,
+    int ms)
 {
 	int  r;
 
@@ -173,6 +176,10 @@ fido_dev_make_cred_wait(fido_dev_t *dev, fido_cred_t *cred, const char *pin, int
 int
 fido_dev_make_cred(fido_dev_t *dev, fido_cred_t *cred, const char *pin)
 {
+#ifdef USE_WINHELLO
+	if (dev->flags & FIDO_DEV_WINHELLO)
+		return (fido_winhello_make_cred(dev, cred, pin));
+#endif
 	if (fido_dev_is_fido2(dev) == false) {
 		if (pin != NULL || cred->rk == FIDO_OPT_TRUE ||
 		    cred->ext.mask != 0)
@@ -184,9 +191,16 @@ fido_dev_make_cred(fido_dev_t *dev, fido_cred_t *cred, const char *pin)
 }
 
 static int
-check_extensions(const fido_cred_ext_t *authdata_ext, const fido_cred_ext_t *ext)
+check_extensions(const fido_cred_ext_t *authdata_ext,
+    const fido_cred_ext_t *ext)
 {
-	return (timingsafe_bcmp(authdata_ext, ext, sizeof(*authdata_ext)));
+	fido_cred_ext_t	 tmp;
+
+	/* XXX: largeBlobKey is not part of the extensions map */
+	memcpy(&tmp, ext, sizeof(tmp));
+	tmp.mask &= ~FIDO_EXT_LARGEBLOB_KEY;
+
+	return (timingsafe_bcmp(authdata_ext, &tmp, sizeof(*authdata_ext)));
 }
 
 int
@@ -326,7 +340,7 @@ fido_cred_verify(const fido_cred_t *cred)
 			r = FIDO_ERR_INTERNAL;
 			goto out;
 		}
-	} else {
+	} else if (!strcmp(cred->fmt, "fido-u2f")) {
 		if (get_signed_hash_u2f(&dgst, cred->authdata.rp_id_hash,
 		    sizeof(cred->authdata.rp_id_hash), &cred->cdh,
 		    &cred->attcred.id, &cred->attcred.pubkey.es256) < 0) {
@@ -334,6 +348,10 @@ fido_cred_verify(const fido_cred_t *cred)
 			r = FIDO_ERR_INTERNAL;
 			goto out;
 		}
+	} else {
+		fido_log_debug("%s: unknown fmt %s", __func__, cred->fmt);
+		r = FIDO_ERR_INVALID_ARGUMENT;
+		goto out;
 	}
 
 	if (verify_sig(&dgst, &cred->attstmt.x5c, &cred->attstmt.sig) < 0) {
@@ -401,7 +419,7 @@ fido_cred_verify_self(const fido_cred_t *cred)
 			r = FIDO_ERR_INTERNAL;
 			goto out;
 		}
-	} else {
+	} else if (!strcmp(cred->fmt, "fido-u2f")) {
 		if (get_signed_hash_u2f(&dgst, cred->authdata.rp_id_hash,
 		    sizeof(cred->authdata.rp_id_hash), &cred->cdh,
 		    &cred->attcred.id, &cred->attcred.pubkey.es256) < 0) {
@@ -409,6 +427,10 @@ fido_cred_verify_self(const fido_cred_t *cred)
 			r = FIDO_ERR_INTERNAL;
 			goto out;
 		}
+	} else {
+		fido_log_debug("%s: unknown fmt %s", __func__, cred->fmt);
+		r = FIDO_ERR_INVALID_ARGUMENT;
+		goto out;
 	}
 
 	switch (cred->attcred.type) {
@@ -451,13 +473,11 @@ fido_cred_new(void)
 static void
 fido_cred_clean_authdata(fido_cred_t *cred)
 {
-	free(cred->authdata_cbor.ptr);
-	free(cred->authdata_raw.ptr);
-	free(cred->attcred.id.ptr);
+	fido_blob_reset(&cred->authdata_cbor);
+	fido_blob_reset(&cred->authdata_raw);
+	fido_blob_reset(&cred->attcred.id);
 
 	memset(&cred->authdata_ext, 0, sizeof(cred->authdata_ext));
-	memset(&cred->authdata_cbor, 0, sizeof(cred->authdata_cbor));
-	memset(&cred->authdata_raw, 0, sizeof(cred->authdata_raw));
 	memset(&cred->authdata, 0, sizeof(cred->authdata));
 	memset(&cred->attcred, 0, sizeof(cred->attcred));
 }
@@ -465,16 +485,18 @@ fido_cred_clean_authdata(fido_cred_t *cred)
 void
 fido_cred_reset_tx(fido_cred_t *cred)
 {
-	free(cred->cdh.ptr);
+	fido_blob_reset(&cred->cd);
+	fido_blob_reset(&cred->cdh);
+	fido_blob_reset(&cred->user.id);
+	fido_blob_reset(&cred->blob);
+
 	free(cred->rp.id);
 	free(cred->rp.name);
-	free(cred->user.id.ptr);
 	free(cred->user.icon);
 	free(cred->user.name);
 	free(cred->user.display_name);
 	fido_free_blob_array(&cred->excl);
 
-	memset(&cred->cdh, 0, sizeof(cred->cdh));
 	memset(&cred->rp, 0, sizeof(cred->rp));
 	memset(&cred->user, 0, sizeof(cred->user));
 	memset(&cred->excl, 0, sizeof(cred->excl));
@@ -485,31 +507,15 @@ fido_cred_reset_tx(fido_cred_t *cred)
 	cred->uv = FIDO_OPT_OMIT;
 }
 
-static void
-fido_cred_clean_x509(fido_cred_t *cred)
-{
-	free(cred->attstmt.x5c.ptr);
-	cred->attstmt.x5c.ptr = NULL;
-	cred->attstmt.x5c.len = 0;
-}
-
-static void
-fido_cred_clean_sig(fido_cred_t *cred)
-{
-	free(cred->attstmt.sig.ptr);
-	cred->attstmt.sig.ptr = NULL;
-	cred->attstmt.sig.len = 0;
-}
-
 void
 fido_cred_reset_rx(fido_cred_t *cred)
 {
 	free(cred->fmt);
 	cred->fmt = NULL;
-
 	fido_cred_clean_authdata(cred);
-	fido_cred_clean_x509(cred);
-	fido_cred_clean_sig(cred);
+	fido_blob_reset(&cred->attstmt.x5c);
+	fido_blob_reset(&cred->attstmt.sig);
+	fido_blob_reset(&cred->largeblob_key);
 }
 
 void
@@ -519,12 +525,9 @@ fido_cred_free(fido_cred_t **cred_p)
 
 	if (cred_p == NULL || (cred = *cred_p) == NULL)
 		return;
-
 	fido_cred_reset_tx(cred);
 	fido_cred_reset_rx(cred);
-
 	free(cred);
-
 	*cred_p = NULL;
 }
 
@@ -610,21 +613,20 @@ fail:
 
 }
 
+int
+fido_cred_set_id(fido_cred_t *cred, const unsigned char *ptr, size_t len)
+{
+	if (fido_blob_set(&cred->attcred.id, ptr, len) < 0)
+		return (FIDO_ERR_INVALID_ARGUMENT);
+
+	return (FIDO_OK);
+}
+
 int
 fido_cred_set_x509(fido_cred_t *cred, const unsigned char *ptr, size_t len)
 {
-	unsigned char *x509;
-
-	fido_cred_clean_x509(cred);
-
-	if (ptr == NULL || len == 0)
+	if (fido_blob_set(&cred->attstmt.x5c, ptr, len) < 0)
 		return (FIDO_ERR_INVALID_ARGUMENT);
-	if ((x509 = malloc(len)) == NULL)
-		return (FIDO_ERR_INTERNAL);
-
-	memcpy(x509, ptr, len);
-	cred->attstmt.x5c.ptr = x509;
-	cred->attstmt.x5c.len = len;
 
 	return (FIDO_OK);
 }
@@ -632,18 +634,8 @@ fido_cred_set_x509(fido_cred_t *cred, const unsigned char *ptr, size_t len)
 int
 fido_cred_set_sig(fido_cred_t *cred, const unsigned char *ptr, size_t len)
 {
-	unsigned char *sig;
-
-	fido_cred_clean_sig(cred);
-
-	if (ptr == NULL || len == 0)
+	if (fido_blob_set(&cred->attstmt.sig, ptr, len) < 0)
 		return (FIDO_ERR_INVALID_ARGUMENT);
-	if ((sig = malloc(len)) == NULL)
-		return (FIDO_ERR_INTERNAL);
-
-	memcpy(sig, ptr, len);
-	cred->attstmt.sig.ptr = sig;
-	cred->attstmt.sig.len = len;
 
 	return (FIDO_OK);
 }
@@ -676,11 +668,28 @@ fido_cred_exclude(fido_cred_t *cred, const unsigned char *id_ptr, size_t id_len)
 	return (FIDO_OK);
 }
 
+int
+fido_cred_set_clientdata(fido_cred_t *cred, const unsigned char *data,
+    size_t data_len)
+{
+	if (!fido_blob_is_empty(&cred->cdh) ||
+	    fido_blob_set(&cred->cd, data, data_len) < 0) {
+		return (FIDO_ERR_INVALID_ARGUMENT);
+	}
+	if (fido_sha256(&cred->cdh, data, data_len) < 0) {
+		fido_blob_reset(&cred->cd);
+		return (FIDO_ERR_INTERNAL);
+	}
+
+	return (FIDO_OK);
+}
+
 int
 fido_cred_set_clientdata_hash(fido_cred_t *cred, const unsigned char *hash,
     size_t hash_len)
 {
-	if (fido_blob_set(&cred->cdh, hash, hash_len) < 0)
+	if (!fido_blob_is_empty(&cred->cd) ||
+	    fido_blob_set(&cred->cdh, hash, hash_len) < 0)
 		return (FIDO_ERR_INVALID_ARGUMENT);
 
 	return (FIDO_OK);
@@ -740,12 +749,8 @@ fido_cred_set_user(fido_cred_t *cred, const unsigned char *user_id,
 		up->icon = NULL;
 	}
 
-	if (user_id != NULL) {
-		if ((up->id.ptr = malloc(user_id_len)) == NULL)
-			goto fail;
-		memcpy(up->id.ptr, user_id, user_id_len);
-		up->id.len = user_id_len;
-	}
+	if (user_id != NULL && fido_blob_set(&up->id, user_id, user_id_len) < 0)
+		goto fail;
 	if (name != NULL && (up->name = strdup(name)) == NULL)
 		goto fail;
 	if (display_name != NULL &&
@@ -776,8 +781,7 @@ fido_cred_set_extensions(fido_cred_t *cred, int ext)
 	if (ext == 0)
 		cred->ext.mask = 0;
 	else {
-		if (ext != FIDO_EXT_HMAC_SECRET &&
-		    ext != FIDO_EXT_CRED_PROTECT)
+		if ((ext & FIDO_EXT_CRED_MASK) != ext)
 			return (FIDO_ERR_INVALID_ARGUMENT);
 		cred->ext.mask |= ext;
 	}
@@ -829,6 +833,19 @@ fido_cred_set_prot(fido_cred_t *cred, int prot)
 	return (FIDO_OK);
 }
 
+int
+fido_cred_set_blob(fido_cred_t *cred, const unsigned char *ptr, size_t len)
+{
+	if (ptr == NULL || len == 0)
+		return (FIDO_ERR_INVALID_ARGUMENT);
+	if (fido_blob_set(&cred->blob, ptr, len) < 0)
+		return (FIDO_ERR_INTERNAL);
+
+	cred->ext.mask |= FIDO_EXT_CRED_BLOB;
+
+	return (FIDO_OK);
+}
+
 int
 fido_cred_set_fmt(fido_cred_t *cred, const char *fmt)
 {
@@ -838,7 +855,8 @@ fido_cred_set_fmt(fido_cred_t *cred, const char *fmt)
 	if (fmt == NULL)
 		return (FIDO_ERR_INVALID_ARGUMENT);
 
-	if (strcmp(fmt, "packed") && strcmp(fmt, "fido-u2f"))
+	if (strcmp(fmt, "packed") && strcmp(fmt, "fido-u2f") &&
+	    strcmp(fmt, "none"))
 		return (FIDO_ERR_INVALID_ARGUMENT);
 
 	if ((cred->fmt = strdup(fmt)) == NULL)
@@ -1054,3 +1072,15 @@ fido_cred_user_id_len(const fido_cred_t *cred)
 {
 	return (cred->user.id.len);
 }
+
+const unsigned char *
+fido_cred_largeblob_key_ptr(const fido_cred_t *cred)
+{
+	return (cred->largeblob_key.ptr);
+}
+
+size_t
+fido_cred_largeblob_key_len(const fido_cred_t *cred)
+{
+	return (cred->largeblob_key.len);
+}
